#include <stdio.h>
#include <zephyr.h>
#include <kernel.h>
#include <arch/cpu.h>
#include <sys/printk.h>
//#include <logging/log.h>
#include <settings/settings.h>

#include <cstring>
#include "nrf.h"
#include <nrfx_rtc.h>
#include <string.h>
#include <zephyr/pm/pm.h>
#include <zephyr/pm/policy.h>
#include <zephyr/pm/device.h>

//#include <stats/stats.h>
//#include <mgmt/buf.h>

#ifdef CONFIG_MCUMGR_CMD_FS_MGMT
#include <device.h>
#include <fs/fs.h>
#include "fs_mgmt/fs_mgmt.h"
#include <fs/littlefs.h>
#endif
#ifdef CONFIG_MCUMGR_CMD_OS_MGMT
#include "os_mgmt/os_mgmt.h"
#endif
#ifdef CONFIG_MCUMGR_CMD_IMG_MGMT
#include "img_mgmt/img_mgmt.h"
#endif
#ifdef CONFIG_MCUMGR_CMD_STAT_MGMT
#include "stat_mgmt/stat_mgmt.h"
#endif

#ifdef CONFIG_MCUMGR_SMP_BT
#include <bluetooth/bluetooth.h>
#include <bluetooth/conn.h>
#include <bluetooth/gatt.h>
#include <mgmt/mcumgr/smp_bt.h>
#endif

/**********SPIM Includes*******************/
#include "nrf.h"
#include <nrfx_spim.h>
#include <nrfx_gpiote.h>
#include <drivers/nrfx_errors.h>

#define WAKEUP_PIN          NRF_GPIO_PIN_MAP(1,13)
#define TEST_GPIO	        NRF_GPIO_PIN_MAP(0,31)
#define APP_SPIM_SCK_PIN    (36)
#define APP_SPIM_MOSI_PIN   (37)
#define APP_SPIM_MISO_PIN   (38)
#define APP_SPIM_nCS   	    NRF_GPIO_PIN_MAP(1,12)

// SPI Interface functions for Nordic
#define SPI_INSTANCE2 2
#define SPI_INSTANCE4 4

static const nrfx_spim_t spim = NRFX_SPIM_INSTANCE(SPI_INSTANCE2);
static const nrfx_spim_t spim4 = NRFX_SPIM_INSTANCE(SPI_INSTANCE4);

static volatile bool spi_xfer_done;    /**< Flag used    to indicate that SPI instance completed the transfer. */

void wakeup_detected(nrfx_gpiote_pin_t pin, nrf_gpiote_polarity_t action)
{
    struct pm_state_info wake_state = {PM_STATE_ACTIVE, 0, 0};
    //pm_state_set(PM_STATE_ACTIVE, PM_STATE_ACTIVE);
    pm_state_force(0u, &wake_state);
}

// Configurations
// Method to configure gpio
void vhi_gpio_init(void)
{
 	int32_t error;
	const nrfx_gpiote_out_config_t out_config = NRFX_GPIOTE_CONFIG_OUT_SIMPLE(false);
	const nrfx_gpiote_in_config_t in_config = {
		.sense = NRF_GPIOTE_POLARITY_HITOLO,
		.pull = NRF_GPIO_PIN_NOPULL,
		.is_watcher = false,
		.hi_accuracy = true,
		.skip_gpio_setup = false,
	};

	IRQ_CONNECT(GPIOTE1_IRQn, 4, nrfx_isr, nrfx_gpiote_irq_handler, 0);
	IRQ_CONNECT(GPIOTE0_IRQn, 4, nrfx_isr, nrfx_gpiote_irq_handler, 0);

	error = nrfx_gpiote_init(0);
	if (error != NRFX_SUCCESS)
	{
		printk("nrfx_gpiote_init error: %08x", error);
		return;
	}

    error = nrfx_gpiote_out_init(TEST_GPIO, &out_config);
	if (error != NRFX_SUCCESS) {
		printk("nrfx_gpiote_out_init error: %08x", error);
        return;
    }

	// Setup magnet on the board for nrfx_gpiote
    error = nrfx_gpiote_in_init(WAKEUP_PIN, &in_config, wakeup_detected);
	if (error != NRFX_SUCCESS)
	{
		printk("Cannot init magnet (error: %d)\n", error);
	}
    nrfx_gpiote_in_event_enable(WAKEUP_PIN, true);
//DEVICE_DT_GET(DT_NODELABEL(gpiote0));  causes build error
    // const struct device *gpio1_dev = device_get_binding("GPIOTE_0");
    // if (pm_device_wakeup_is_capable(gpio1_dev) == true)
    // {
    //     printk("GPIO1 as wakeup capable\n");
    // }
    // if (pm_device_wakeup_enable(gpio1_dev, true) == false)
    // {
    //     printk("Cannot set GPIO1 as wakeup\n");
    // }

    nrfx_gpiote_out_set(TEST_GPIO);
}

#ifdef CONFIG_MCUMGR_CMD_FS_MGMT
FS_LITTLEFS_DECLARE_DEFAULT_CONFIG(cstorage);
static struct fs_mount_t littlefs_mnt = {
	.type = FS_LITTLEFS,
	.mnt_point = "/lfs",
	.fs_data = &cstorage,
	.storage_dev = (void *)FLASH_AREA_ID(storage)
};
#endif
/* C:\ncs\v2.0.0\zephyr\samples\boards\nrf\system_off\src
 * Prevent deep sleep (system off) from being entered on long timeouts
 * or `K_FOREVER` due to the default residency policy.
 *
 * This has to be done before anything tries to sleep, which means
 * before the threading system starts up between PRE_KERNEL_2 and
 * POST_KERNEL.  Do it at the start of PRE_KERNEL_2.
 */
static int disable_ds_1(const struct device *dev)
{
	ARG_UNUSED(dev);

   	/* Prevent deep sleep (system off) from being entered */
	pm_policy_state_lock_get(PM_STATE_SOFT_OFF);
	return 0;
}

SYS_INIT(disable_ds_1, PRE_KERNEL_2, 0);

void spim_event_handler(nrfx_spim_evt_t const *p_event, void *p_context)
{
    spi_xfer_done = true;
}

static void spim_initialize() {
    nrfx_err_t err;
    nrfx_spim_config_t spim_config =
	NRFX_SPIM_DEFAULT_CONFIG(APP_SPIM_SCK_PIN, APP_SPIM_MOSI_PIN, APP_SPIM_MISO_PIN, NRFX_SPIM_PIN_NOT_USED);
    spim_config.frequency = NRF_SPIM_FREQ_4M;
    spim_config.mode = NRF_SPIM_MODE_0;
    spim_config.bit_order = NRF_SPIM_BIT_ORDER_MSB_FIRST;
    spim_config.miso_pull = NRF_GPIO_PIN_NOPULL;
    spim_config.irq_priority = NRFX_SPIM_DEFAULT_CONFIG_IRQ_PRIORITY;

	err = nrfx_spim_init(&spim, &spim_config, spim_event_handler, NULL);
	if (err != NRFX_SUCCESS) {
		printk("nrfx_spim_init() failed: 0x%08x\n", err);
	}
}

void app_spim_pin_config(void) {

    nrfx_err_t err_code;

    const nrfx_gpiote_out_config_t out_config = NRFX_GPIOTE_CONFIG_OUT_SIMPLE(false);

    err_code = nrfx_gpiote_out_init(APP_SPIM_nCS, &out_config);
	if (err_code != NRFX_SUCCESS) {
        return;
    }
}
const struct device *spi2_dev;

void APP_spim_initialize() {
    spi2_dev = device_get_binding("SPI_2");
    // init the reset and interrupt pins
    app_spim_pin_config();

	IRQ_CONNECT(DT_IRQN(DT_NODELABEL(spi2)),
				DT_IRQ(DT_NODELABEL(spi2), priority),
				nrfx_isr, nrfx_spim_2_irq_handler, 0);

    spim_initialize();
    nrfx_gpiote_out_set(APP_SPIM_nCS);
    pm_device_action_run(spi2_dev, PM_DEVICE_ACTION_SUSPEND);
}

void APP_spim_on() {
    pm_device_action_run(spi2_dev, PM_DEVICE_ACTION_RESUME);
    nrfx_gpiote_out_uninit(APP_SPIM_SCK_PIN);
    nrfx_gpiote_out_uninit(APP_SPIM_MOSI_PIN);
    nrfx_gpiote_out_uninit(APP_SPIM_MISO_PIN);
    spim_initialize();
}

void APP_spim_off() {
    nrfx_gpiote_out_clear(APP_SPIM_nCS);;

    pm_device_action_run(spi2_dev, PM_DEVICE_ACTION_SUSPEND);
    nrfx_spim_uninit(&spim);
    const nrfx_gpiote_out_config_t out_config = NRFX_GPIOTE_CONFIG_OUT_SIMPLE(false);

    nrfx_err_t err_code = nrfx_gpiote_out_init(APP_SPIM_SCK_PIN, &out_config);
    if (err_code != NRFX_SUCCESS) {
        return;
    }

    err_code = nrfx_gpiote_out_init(APP_SPIM_MOSI_PIN, &out_config);
    if (err_code != NRFX_SUCCESS) {
        return;
    }

    err_code = nrfx_gpiote_out_init(APP_SPIM_MISO_PIN, &out_config);
    if (err_code != NRFX_SUCCESS) {
        return;
    }
}

typedef enum
{
    PowerOnInit,
    Polling,
    IdleWait,
    DeepSleep
} power_state_enum_t;

power_state_enum_t power_state;

int main(void)
{
	/* Initialize GPIO */
	vhi_gpio_init();

	// This MUST be called before any ACC functions
	APP_spim_initialize();
    /// BLE config removed

    while (1)
    {
        switch (power_state)
        {
            case PowerOnInit:
                APP_spim_on();
                //SPI device Init here
                power_state = Polling;
                break;

            case Polling:
                // simulate SPI device Polling here
                for (int i = 0; i < 400000; i++)
                {
                    __NOP();
                }

                APP_spim_off();
                // SPI device un-initialize here
                power_state = IdleWait;
                break;

            case IdleWait:
                // Simulate sleep SUSPEND_TO_RAM here
                k_sleep(K_MSEC(5000));
                power_state = PowerOnInit;
                break;

            case DeepSleep:
            {
                // Force deep sleep when requested
                struct pm_state_info sleep_state = {PM_STATE_SOFT_OFF, 0, 0};
                pm_state_force(0u, &sleep_state);
                k_cpu_idle();
            }
        }
    }

	return 0;
}
